/*

ftpii -- an FTP server for the Wii

Copyright (C) 2008 Joseph Jordan <joe.ftpii@psychlaw.com.au>

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from
the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1.The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be
appreciated but is not required.

2.Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

3.This notice may not be removed or altered from any source distribution.

*/
#include <errno.h>
#include <network.h>
#include <ogcsys.h>
#include <stdio.h>
#include <string.h>
#include <sys/dir.h>
#include <wiiuse/wpad.h>
#include <unistd.h>

#include "common.h"

static const u16 SRC_PORT = 20;
static const u16 PORT = 21;
static const s32 EQUIT = 696969;

// state associated with client
static char client_representation_type = 'A'; // XXX: ASCII and binary are currently identical...STOR doesn't like this
static struct sockaddr_in client_address;
static s32 passive_socket = -1;

extern u32 net_gethostip();

/*
    TODO: support multi-line reply
*/
static s32 write_reply(s32 peer, u16 code, char *msg) {
    u32 msglen = 4 + strlen(msg) + CRLF_LENGTH;
    char msgbuf[msglen + 1];
    if (msgbuf == NULL) return -ENOMEM;
    sprintf(msgbuf, "%u %s\r\n", code, msg);
    printf("Wrote reply: %s", msgbuf);
    return write_exact(peer, msgbuf, msglen);
}

static s32 ftp_USER(s32 peer, char *username) {
    return write_reply(peer, 331, "User name okay, need password.");
}

static s32 ftp_PASS(s32 peer, char *password) {
    return write_reply(peer, 230, "User logged in, proceed.");
}

static s32 ftp_REIN(s32 peer, char *rest) {
    if (chdir("/")) {
        return write_reply(peer, 550, "Could not change to root directory, connection not reinitialised.");
    }
    client_representation_type = 'A';
    if (passive_socket >= 0) {
        net_close(passive_socket);
        passive_socket = -1;
    }
    return write_reply(peer, 220, "Service ready for new user.");
}

static s32 ftp_QUIT(s32 peer, char *rest) {
    // TODO: dont quit if xfer in progress
    s32 result = write_reply(peer, 221, "Service closing control connection.");
    return result < 0 ? result : -EQUIT;
}

static s32 ftp_SYST(s32 peer, char *rest) {
    return write_reply(peer, 215, "UNIX Type: L8 Version: ftpii");
}

static s32 ftp_TYPE(s32 peer, char *rest) {
    char representation_type[BUFFER_SIZE], param[BUFFER_SIZE];
    char *args[] = { representation_type, param };
    u32 num_args = split(rest, ' ', 1, args);
    if (num_args == 0) {
        return write_reply(peer, 501, "Syntax error in parameters or arguments.");
    } else if ((!strcasecmp("A", representation_type) && (!*param || !strcasecmp("N", param))) ||
               (!strcasecmp("I", representation_type) && num_args == 1)) {
        client_representation_type = *representation_type;
    } else {
        return write_reply(peer, 501, "Syntax error in parameters or arguments.");
    }
    char msg[15];
    sprintf(msg, "Type set to %s.", representation_type);
    return write_reply(peer, 200, msg);
}

static s32 ftp_MODE(s32 peer, char *rest) {
    if (!strcasecmp("S", rest)) {
        return write_reply(peer, 200, "Mode S ok.");
    } else {
        return write_reply(peer, 501, "Syntax error in parameters or arguments.");
    }
}

static s32 ftp_PWD(s32 peer, char *rest) {
    char path[MAXPATHLEN];
    if (getcwd(path, MAXPATHLEN)) {
        // TODO: escape double-quotes
        char msg[MAXPATHLEN + 24];
        // XXX: for now, strip "fat:" from all paths we send to the client, because it screws with some FTP clients (e.g. FileZilla)
        char *stripped_path = path;
        if (strncmp(path, "fat:", 4) == 0) stripped_path += 4;
        sprintf(msg, "\"%s\" is current directory.", stripped_path);
        return write_reply(peer, 257, msg);
    } else {
        return write_reply(peer, 451, path);
    }
}

static s32 ftp_CWD(s32 peer, char *path) {
    struct stat st;
    if (stat(path, &st)) { // have to check this because if we give chdir bad input (e.g. "/fat:/SNES9X/") it can cause a crash
        return write_reply(peer, 550, strerror(errno));
    }
    if (!chdir(path)) {
        return write_reply(peer, 250, "CWD command successful.");
    } else  {
        return write_reply(peer, 550, strerror(errno));
    }
}

static s32 ftp_CDUP(s32 peer, char *rest) {
    if (!chdir("..")) {
        return write_reply(peer, 250, "CDUP command successful.");
    } else  {
        return write_reply(peer, 550, strerror(errno));
    }
}

static s32 ftp_DELE(s32 peer, char *path) {
    if (!unlink(path)) {
        return write_reply(peer, 250, "File or directory removed.");
    } else {
        return write_reply(peer, 550, strerror(errno));
    }
}

static s32 ftp_MKD(s32 peer, char *path) {
    if (strlen(path) == 0) {
        return write_reply(peer, 501, "Syntax error in parameters or arguments.");
    }
    if (!mkdir(path, 0777)) {
        // TODO: error-handling =P
        char old_cwd[MAXPATHLEN];
        char new_path[MAXPATHLEN];
        getcwd(old_cwd, MAXPATHLEN);
        chdir(path);
        getcwd(new_path, MAXPATHLEN);
        // TODO: escape double-quotes
        chdir(old_cwd);
        char msg[MAXPATHLEN + 21];
        // XXX: for now, strip "fat:" from all paths we send to the client, because it screws with some FTP clients (e.g. FileZilla)
        char *stripped_path = new_path;
        if (strncmp(new_path, "fat:", 4) == 0) stripped_path += 4;
        sprintf(msg, "\"%s\" directory created.", stripped_path);
        return write_reply(peer, 257, msg);
    } else {
        return write_reply(peer, 550, strerror(errno));
    }
}

static s32 ftp_SIZE(s32 peer, char *path) {
    struct stat st;
    if (!stat(path, &st)) {
        char size_buf[12];
        sprintf(size_buf, "%li", st.st_size); // XXX: what does this do for files over 2GB?
        return write_reply(peer, 213, size_buf);
    } else {
        return write_reply(peer, 550, strerror(errno));
    }
}

static s32 ftp_PASV(s32 peer, char *rest) {
    if (passive_socket >= 0) net_close(passive_socket);
    passive_socket = net_socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (passive_socket < 0) {
        return write_reply(peer, 520, "Unable to create listening socket.");
    }
    struct sockaddr_in bindAddress;
    memset(&bindAddress, 0, sizeof(bindAddress));
    bindAddress.sin_family = AF_INET;
    bindAddress.sin_port = htons(1024);
    bindAddress.sin_addr.s_addr = htonl(INADDR_ANY);
    s32 result;
    if ((result = net_bind(passive_socket, (struct sockaddr *)&bindAddress, sizeof(bindAddress))) < 0) {
        net_close(passive_socket);
        passive_socket = -1;
        return write_reply(peer, 520, "Unable to bind listening socket.");
    }
    if ((result = net_listen(passive_socket, 1)) < 0) {
        net_close(passive_socket);
        passive_socket = -1;
        return write_reply(peer, 520, "Unable to listen on socket.");
    }
    char reply[49];
    u16 port = bindAddress.sin_port;
    u32 ip = net_gethostip();
    printf("Listening for data connections at %s:%u...\n", inet_ntoa(*(struct in_addr *)&ip), port);
    sprintf(reply, "Entering Passive Mode (%u,%u,%u,%u,%u,%u).", (ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff, ip & 0xff, (port >> 8) & 0xff, port & 0xff);
    return write_reply(peer, 227, reply);
}

typedef s32 (*data_connection_callback_type)(s32 data_socket, void *arg);
typedef s32 (*data_connection_handler_type)(s32 peer, data_connection_callback_type callback, void *arg);

static s32 do_data_connection_active(s32 peer, data_connection_callback_type callback, void *arg) {
    s32 data_socket = net_socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    if (data_socket < 0) {
        printf("DEBUG: Unable to create data socket: [%i] %s\n", -data_socket, strerror(-data_socket));
        return data_socket;
    }
    printf("Attempting to connect to client at %s:%u\n", inet_ntoa(client_address.sin_addr), ntohs(client_address.sin_port));
    struct sockaddr_in bindAddress;
    memset(&bindAddress, 0, sizeof(bindAddress));
    bindAddress.sin_family = AF_INET;
    bindAddress.sin_port = htons(SRC_PORT);
    bindAddress.sin_addr.s_addr = htonl(INADDR_ANY);
    s32 result;
    if ((result = net_bind(data_socket, (struct sockaddr *)&bindAddress, sizeof(bindAddress))) < 0) {
        printf("DEBUG: Unable to bind data socket: [%i] %s\n", -result, strerror(-result));
        net_close(data_socket);
        return result;
    }
    if ((result = net_connect(data_socket, (struct sockaddr *)&client_address, sizeof(client_address))) < 0) {
        printf("Unable to connect to client: [%i] %s\n", -result, strerror(-result));
        net_close(data_socket);
        return result;
    }
    printf("Connected to client!  Transferring data...\n");

    result = callback(data_socket, arg);

    net_close(data_socket);
    if (result < 0) {
        printf("Error occurred while sending data...\n");
    } else {
        printf("Finished transferring data!\n");
    }
    return result;
}

static s32 do_data_connection_passive(s32 peer, data_connection_callback_type callback, void *arg) {
    struct sockaddr_in data_peer_address;
    socklen_t addrlen = sizeof(data_peer_address);
    printf("Waiting for data connections...\n");
    s32 data_socket = net_accept(passive_socket, (struct sockaddr *)&data_peer_address ,&addrlen);
    if (data_socket < 0) {
        printf("DEBUG: Unable to accept data socket: [%i] %s\n", -data_socket, strerror(-data_socket));
        return data_socket;
    }
    printf("Connected to client!  Transferring data...\n");

    s32 result = callback(data_socket, arg);

    net_close(data_socket);
    if (result < 0) {
        printf("Error occurred while sending data...\n");
    } else {
        printf("Finished transferring data!\n");
    }
    return result;
}

static s32 do_data_connection(s32 peer, data_connection_callback_type callback, void *arg) {
    data_connection_handler_type handler = do_data_connection_active;
    if (passive_socket >= 0) handler = do_data_connection_passive;
    return handler(peer, callback, arg);
}

static s32 send_nlst(s32 data_socket, DIR_ITER *dir) {
    s32 result = 0;
    char filename[MAXPATHLEN + 2];
    struct stat st;
	while (dirnext(dir, filename, &st) == 0) {
        size_t end_index = strlen(filename);
        filename[end_index] = CRLF[0];
        filename[end_index + 1] = CRLF[1];
        filename[end_index + 2] = '\0';
        if ((result = write_exact(data_socket, filename, strlen(filename))) < 0) {
            break;
        }
	}
    return result;
}

static s32 send_list(s32 data_socket, DIR_ITER *dir) {
    s32 result = 0;
    char filename[MAXPATHLEN];
    struct stat st;
    char line[MAXPATHLEN + 56 + CRLF_LENGTH + 1];
	while (dirnext(dir, filename, &st) == 0) {
        sprintf(line, "%crwxr-xr-x    1 0        0     %11li Jan 01  1900 %s\r\n", (st.st_mode & S_IFDIR) ? 'd' : '-', st.st_size, filename); // what does it do > 2GB?
        if ((result = write_exact(data_socket, line, strlen(line))) < 0) {
            break;
        }
	}
    return result;
}

static s32 ftp_NLST(s32 peer, char *path) {
    if (strlen(path) == 0) {
        path = ".";
    }
    DIR_ITER *dir = diropen(path);
    if (dir == NULL) {
        return write_reply(peer, 550, strerror(errno));
    }
    s32 result = write_reply(peer, 120, "Writing data.");
    if (result >= 0) {
        result = do_data_connection(peer, (data_connection_callback_type)send_nlst, dir);
        if (result < 0) {
            result = write_reply(peer, 520, "Closing data connection, error occurred during transfer.");
        } else {
            result = write_reply(peer, 226, "Closing data connection, transfer successful.");
        }
    }
    dirclose(dir);
    return result;
}

static s32 ftp_LIST(s32 peer, char *path) {
    if (strlen(path) == 0) {
        path = ".";
    }
    DIR_ITER *dir = diropen(path);
    if (dir == NULL) {
        return write_reply(peer, 550, strerror(errno));
    }
    s32 result = write_reply(peer, 120, "Writing data.");
    if (result >= 0) {
        result = do_data_connection(peer, (data_connection_callback_type)send_list, dir);
        if (result < 0) {
            result = write_reply(peer, 520, "Closing data connection, error occurred during transfer.");
        } else {
            result = write_reply(peer, 226, "Closing data connection, transfer successful.");
        }
    }
    dirclose(dir);
    return result;
}

static s32 ftp_RETR(s32 peer, char *path) {
    FILE *f = fopen(path, "rb");
    if (!f) {
        return write_reply(peer, 550, strerror(errno));
    }
    s32 result = write_reply(peer, 150, "File status okay; writing data.");
    if (result >= 0) {
        result = do_data_connection(peer, (data_connection_callback_type)write_from_file, f);
        if (result < 0) {
            result = write_reply(peer, 520, "Closing data connection, error occurred during transfer.");
        } else {
            result = write_reply(peer, 226, "Closing data connection, transfer successful.");
        }
    }
    fclose(f);
    return result;
}

static s32 ftp_STOR(s32 peer, char *path) {
    FILE *f = fopen(path, "wb");
    if (!f) {
        return write_reply(peer, 550, strerror(errno));
    }
    s32 result = write_reply(peer, 150, "File status okay; reading data.");
    if (result >= 0) {
        result = do_data_connection(peer, (data_connection_callback_type)read_to_file, f);
        if (result < 0) {
            result = write_reply(peer, 520, "Closing data connection, error occurred during transfer.");
        } else {
            result = write_reply(peer, 226, "Closing data connection, transfer successful.");
        }
    }
    fclose(f);
    return result;
}

static s32 ftp_PORT(s32 peer, char *portspec) {
    u32 h1, h2, h3, h4, p1, p2;
    if (sscanf(portspec, "%3u,%3u,%3u,%3u,%3u,%3u", &h1, &h2, &h3, &h4, &p1, &p2) < 6) {
        return write_reply(peer, 501, "Syntax error in parameters or arguments.");
    }
    char addr_str[44];
    sprintf(addr_str, "%u.%u.%u.%u", h1, h2, h3, h4);
    struct in_addr sin_addr;
    if (!inet_aton(addr_str, &sin_addr)) {
        return write_reply(peer, 501, "Syntax error in parameters or arguments.");
    }
    u16 port = ((p1 &0xff) << 8) | (p2 & 0xff);
    client_address.sin_addr = sin_addr;
    client_address.sin_port = htons(port);
    printf("Set client address to %s:%u\n", addr_str, port);
    if (passive_socket >= 0) {
        net_close(passive_socket);
        passive_socket = -1;
    }
    return write_reply(peer, 200, "PORT command successful.");
}

static s32 ftp_NOOP(s32 peer, char *rest) {
    return write_reply(peer, 200, "NOOP command successful.");
}

static s32 ftp_SUPERFLUOUS(s32 peer, char *rest) {
    return write_reply(peer, 202, "Command not implemented, superfluous at this site.");
}

static s32 ftp_UNKNOWN(s32 peer, char *rest) {
    return write_reply(peer, 502, "Command not implemented.");
}

typedef s32 (*ftp_handler_type)(s32 s, char *args);

/*
    returns negative to signal an error that requires closing the connection
*/
static s32 process_command(s32 peer, char *cmd_line) {
    printf("Got command: %s\n", cmd_line);
    char cmd[BUFFER_SIZE], rest[BUFFER_SIZE];
    char *args[] = { cmd, rest };
    u32 num_args = split(cmd_line, ' ', 1, args); 
    if (num_args == 0) {
        return 0;
    }
    ftp_handler_type handler = ftp_UNKNOWN;
    if      (!strcasecmp("LIST", cmd)) handler = ftp_LIST;
    else if (!strcasecmp("PWD" , cmd)) handler = ftp_PWD;
    else if (!strcasecmp("CWD" , cmd)) handler = ftp_CWD;
    else if (!strcasecmp("CDUP", cmd)) handler = ftp_CDUP;
    else if (!strcasecmp("SIZE", cmd)) handler = ftp_SIZE;
    else if (!strcasecmp("PASV", cmd)) handler = ftp_PASV;
    else if (!strcasecmp("PORT", cmd)) handler = ftp_PORT;
    else if (!strcasecmp("TYPE", cmd)) handler = ftp_TYPE;
    else if (!strcasecmp("SYST", cmd)) handler = ftp_SYST;
    else if (!strcasecmp("MODE", cmd)) handler = ftp_MODE;
    else if (!strcasecmp("RETR", cmd)) handler = ftp_RETR;
    else if (!strcasecmp("STOR", cmd)) handler = ftp_STOR;
    else if (!strcasecmp("DELE", cmd)) handler = ftp_DELE;
    else if (!strcasecmp("MKD",  cmd)) handler = ftp_MKD;
    else if (!strcasecmp("RMD",  cmd)) handler = ftp_DELE;
    else if (!strcasecmp("NLST", cmd)) handler = ftp_NLST;
    else if (!strcasecmp("USER", cmd)) handler = ftp_USER;
    else if (!strcasecmp("PASS", cmd)) handler = ftp_PASS;
    else if (!strcasecmp("QUIT", cmd)) handler = ftp_QUIT;
    else if (!strcasecmp("REIN", cmd)) handler = ftp_REIN;
    else if (!strcasecmp("ALLO", cmd)) handler = ftp_SUPERFLUOUS;
    else if (!strcasecmp("SITE", cmd)) handler = ftp_SUPERFLUOUS;
    else if (!strcasecmp("NOOP", cmd)) handler = ftp_NOOP;
    return handler(peer, rest);
}

static void process_connection(s32 server) {
    s32 peer = accept_peer(server, &client_address);

    if (write_reply(peer, 220, "ftpii") < 0) {
        printf("Error writing greeting.\n");
        net_close(peer);
        return;
    }
    
    char buf[BUFFER_SIZE];
    s32 offset = 0;
    s32 bytes_read;
    while (offset < (BUFFER_SIZE - 1)) {
        char *offset_buf = buf + offset;
        if ((bytes_read = net_read(peer, offset_buf, BUFFER_SIZE - 1 - offset)) < 0) {
            printf("Read error %i occurred, closing client.\n", bytes_read);
            goto recv_loop_end;
        } else if (bytes_read == 0) {
            goto recv_loop_end; // EOF from client
        }
        offset += bytes_read;
        buf[offset] = '\0';
        
        if (strchr(offset_buf, '\0') != (buf + offset)) {
            printf("Received a null byte from client, closing connection ;-)\n"); // i have decided this isn't allowed =P
            goto recv_loop_end;
        }

        char *next;
        char *end;
        for (next = buf; (end = strstr(next, CRLF)); next = end + CRLF_LENGTH) {
            *end = '\0';
            if (strchr(next, '\n')) {
                printf("Received a line-feed from client without carriage return, closing connection ;-)\n"); // i have decided this isn't allowed =P
                goto recv_loop_end;
            }
            
            if (*next) {
                s32 result;
                if ((result = process_command(peer, next)) < 0) {
                    if (result != -EQUIT) {
                        printf("Closing connection due to error while processing command: %s\n", next);
                    }
                    goto recv_loop_end;
                }
            }
            
        }
        
        if (next != buf) { // some lines were processed
            offset = strlen(next);
            char tmp_buf[offset];
            memcpy(tmp_buf, next, offset);
            memcpy(buf, tmp_buf, offset);
        }
    }
    printf("Received line longer than %u bytes, closing client.\n", BUFFER_SIZE - 1);

    recv_loop_end:

    net_close(peer);
    if (chdir("/")) die("Could not change to root directory, exiting");
    client_representation_type = 'A';
    if (passive_socket >= 0) {
        net_close(passive_socket);
        passive_socket = -1;
    }
    printf("Done doing stuffs!\n");
}

static void mainloop() {
    s32 server = create_server(PORT);
    while (1) {
        printf("\nListening on TCP port %u...\n", PORT);
        process_connection(server);
    }
}

int main(int argc, char **argv) {
    initialise_video();
    printf("\n");
    initialise_fat();
    if (chdir("/")) die("Could not change to root directory, exiting");
    WPAD_Init();
    if (initialise_reset_button()) {
        printf("To exit, hold A on WiiMote #1 or press the reset button.\n");
    } else {
        printf("Unable to start reset thread - hold down the power button to exit.\n");
    }
    wait_for_network_initialisation();
    
    mainloop();

    return 0;
}
